#ifndef _PROG_H_
#define _PROG_H_
#include <stdio.h>
#include <ctype.h>
#include <fstream.h>
#include <iostream.h>
#include "defs.h"
#include "video.h"
#include "_tape.h"
#include "utils.h"
#include "state.h"
#include "clock.h"

#define MAX_SYMVOLS  10
#define MAX_LINES  30

char commands[] = {'l', 'r', 's'};

struct TSymvol {
  char symvol;
  int command;
  int state;
};

struct TLine {
  int is_break;
  int state;
  TSymvol symvols[MAX_SYMVOLS];
} lines[MAX_LINES];

class TProgram {
    void Draw();
    char alfabet[10];
    char filename[40];
    int pos_x;
    int pos_y;
    int e1, e2, e3;
    TState *state;
    char cur_sym;
    int cur_move;
    int cur_state;
    long cur_step;
    char attr1, attr2, attr3;
    void UpdateFilename();
  public:
    TProgram();
    ~TProgram();
    void Clear();
    void Update();
    void Edit(unsigned char, char);
    int Run();
    void Reset();
    void GetState();
    void Save();
    void Load();
    void ClearBreakPoints();
};

TProgram::TProgram()
{
  state = new TState(56, 5);
  strcpy(filename, "NONAME.TUB");
  cur_sym = ' ';
  cur_move = 0;
  cur_state = 0;
  cur_step = 0;
  Draw();
  Clear();
}

TProgram::~TProgram()
{
  delete state;
}

void TProgram::Draw()
{
  fill(4, 5, 38, 1, '', PROGRAM_COLOR);
  fill(4, 16, 38, 1, '', PROGRAM_COLOR);
  fill(3, 6, 1, 10, '', PROGRAM_COLOR);
  fill(42, 6, 1, 10, '', PROGRAM_COLOR);
  out_char(3, 5, '', PROGRAM_COLOR);
  out_char(42, 5, '', PROGRAM_COLOR);
  out_char(3, 16, '', PROGRAM_COLOR);
  out_char(42, 16, '', PROGRAM_COLOR);
  fill(4, 6, 38, 10, ' ', PROGRAM_COLOR);
  set_attr(5, 17, 40, 1, SHADOW_COLOR);
  set_attr(43, 6, 2, 12, SHADOW_COLOR);
  fill(4, 7, 38, 1, '', PROGRAM_COLOR);
  for (int i = 0; i < 4; i++)
  {
    out_char(10 + i * 8, 5, '', PROGRAM_COLOR);
    fill(10 + i * 8, 6, 1, 10, '', PROGRAM_COLOR);
    out_char(10 + i * 8, 16, '', PROGRAM_COLOR);
    out_char(10 + i * 8, 7, '', PROGRAM_COLOR);
  }
  out_char(3, 7, '', PROGRAM_COLOR);
  out_char(42, 7, '', PROGRAM_COLOR);
  UpdateFilename();
}

void TProgram::Clear()
{
  pos_x = 0;
  pos_y = 0;
  e1 = 0;
  e2 = 1;
  e3 = 0;
  for (int i = 0; i < MAX_SYMVOLS; i++)
    alfabet[i] = ' ';
  for (int i = 0; i < MAX_LINES; i++)
  {
    lines[i].is_break = 0;
    lines[i].state = i;
    for (int j = 0; j < MAX_SYMVOLS; j++)
    {
      lines[i].symvols[j].symvol = ' ';
      lines[i].symvols[j].command = -1;
      lines[i].symvols[j].state = -1;
    }
  }
  Update();
}

void TProgram::Update()
{
  char buf[10];

  for (int j = 0; j < 4; j++)
    out_char(14 + j * 8, 6, alfabet[j + pos_x], PROGRAM_SYMCOLOR);

  for (int i = 0; i < 8; i++)
  {
    if (lines[i + pos_y].state >= 0)
    {
      sprintf(buf, "%2d", lines[i + pos_y].state);
      out_string(6, i + 8, buf, PROGRAM_STATECOLOR);
    }
    else
      fill(6, i + 8, 2, 1, ' ', PROGRAM_STATECOLOR);
    for (int j = 0; j < 4; j++)
    {
      if ((lines[i + pos_y].symvols[j + pos_x].symvol == alfabet[j + pos_x]) &&
         (lines[i + pos_y].symvols[j + pos_x].command == 2) &&
         (lines[i + pos_y].symvols[j + pos_x].state == lines[i + pos_y].state))
      {
        attr1 = PROGRAM_STOPCOLOR;
        attr2 = attr1;
        attr3 = attr2;
      }
      else
      {
        attr1 = PROGRAM_SYMCOLOR;
        attr2 = PROGRAM_MOVECOLOR;
        attr3 = PROGRAM_STATECOLOR;
      }
      out_char(12 + j * 8, i + 8, lines[i + pos_y].symvols[j + pos_x].symvol, attr1);
      if (lines[i + pos_y].symvols[j + pos_x].command >= 0)
        out_char(14 + j * 8, i + 8,
                 commands[lines[i + pos_y].symvols[j + pos_x].command], attr2);
      else
        fill(14 + j * 8, i + 8, 1, 1, ' ', PROGRAM_MOVECOLOR);
      if (lines[i + pos_y].symvols[j + pos_x].state >= 0)
      {
        sprintf(buf, "%d", lines[i + pos_y].symvols[j + pos_x].state);
        out_string(16 + j * 8, i + 8, buf, attr3);
        if (strlen(buf) < 2)
          out_char(17 + j * 8, i + 8, ' ', attr3);
      }
      else
        fill(16 + j * 8, i + 8, 2, 1, ' ', PROGRAM_STATECOLOR);
    }
    if (lines[i + pos_y].is_break)
    {
      out_char(3, i + 8, '', 0x34);
      out_char(42, i + 8, '', 0x34);
    }
    else
    {
      out_char(3, i + 8, '', PROGRAM_TEXTCOLOR);
      out_char(42, i + 8, '', PROGRAM_TEXTCOLOR);
    }
  }
}

void TProgram::Edit(unsigned char command, char key)
{
  int x, y, a;

  switch (command)
  {
    case 75: if (e1)
             {
               if (!e2)
               {
                 if (pos_x)
                   pos_x--;
               }
               if (e3 > 0)
                 e3--;
               else
               {
                 if (e2 > 0)
                 {
                   e3 = 2;
                   e2--;
                 }
               }
             }
             else
             {
               if (e2 > 1)
               {
                 e3 = 0;
                 e2--;
               }
               else
               if (pos_x)
                 pos_x--;
             }
             break;
    case 77: if (e2 && e1)
             {
               if (e3 < 2)
                 e3++;
               else
               {
                 if (e2 < 4)
                 {
                   e3 = 0;
                   e2++;
                 }
                 else
                 if (pos_x < MAX_SYMVOLS - 4)
                   pos_x++;
               }
             }
             else
             {
               if (e2 < 4)
                 e2++;
               else
               if (pos_x < MAX_SYMVOLS - 4)
                 pos_x++;
               e3 = 0;
             }
             break;
    case 80: if (e1 < 8)
               e1++;
             else
             if (pos_y < MAX_LINES - 8)
               pos_y++;
             break;
    case 72: if (e1 > 1)
               e1--;
             else
             {
               if ((e1 && !e2) || (!e1 && e2))
               {
                 if (pos_y)
                   pos_y--;
               }
               if (e1 && e2)
               {
                 e3 = 0;
                 e1 = 0;
               }
             }
             break;
    case 83: if (e1 && !e2)
             {
               lines[e1 - 1 + pos_y].state = -1;
             }
             else if (e2 && e1)
             {
               switch (e3)
               {
                 case 0: lines[e1 - 1 + pos_y].symvols[e2 - 1 + pos_x].symvol = 0x20; break;
                 case 1: lines[e1 - 1 + pos_y].symvols[e2 - 1 + pos_x].command = -1; break;
                 case 2: lines[e1 - 1 + pos_y].symvols[e2 - 1 + pos_x].state = -1; break;
               }
             }
             else if (!e1)
             {
               alfabet[e2 - 1 + pos_x] = 0x20;
             }
             break;
    case 255: if (e1)
              {
                lines[e1 - 1 + pos_y].is_break ^= 1;
                Update();
                return;
              }
  }

  if (!e1)
  {
    x = (e2 - 1) * 8 + 14;
    y = 6;
    if (key >= 0x20)
      alfabet[e2 - 1 + pos_x] = key;
  }
  else
  {
    if (!e2)
    {
      x = 6;
      y = e1 + 7;
      if ((key >= 0x30) && (key <= 0x39))
      {
        a = lines[e1 - 1 + pos_y].state;
        if (a < 0)
          a = key - 0x30;
        else
        {
          if (a > 9)
            a = 0;
          a *= 10;
          a += key - 0x30;
        }
        lines[e1 - 1 + pos_y].state = a;
      }
      if (key == 0x20)
        lines[e1 - 1 + pos_y].state = -1;
    }
    else
    {
      x = (e2 - 1) * 8 + 2 * e3 + 12;
      y = e1 + 7;
      switch (e3)
      {
        case 0: if (key >= 0x20)
                  lines[e1 - 1 + pos_y].symvols[e2 - 1 + pos_x].symvol = key;
                break;
        case 1: switch (key)
                {
                  case 'l': lines[e1 - 1 + pos_y].symvols[e2 - 1 + pos_x].command = 0; break;
                  case 'r': lines[e1 - 1 + pos_y].symvols[e2 - 1 + pos_x].command = 1; break;
                  case 's': lines[e1 - 1 + pos_y].symvols[e2 - 1 + pos_x].command = 2; break;
                }
                break;
        case 2: a = lines[e1 - 1 + pos_y].symvols[e2 - 1 + pos_x].state;
                if ((key >= 0x30) && (key <= 0x39))
                {
                  if (a < 0)
                    a = key - 0x30;
                  else
                  {
                    if (a > 9)
                      a = 0;
                    a *= 10;
                    a += key - 0x30;
                  }
                }
                else
                if (key == 0x20)
                  a = -1;
                lines[e1 - 1 + pos_y].symvols[e2 - 1 + pos_x].state = a;
      }
    }
  }
  gotoxy(x + 1, y + 1);

  Update();
  set_attr(x, y, 1, 1, 0x1B);
}

int TProgram::Run()
{
  char t, error_str[80];
  int s, err_l, err_c;
  cur_sym = tape -> GetData();
  t = cur_sym;
  s = cur_state;
  for (int i = 0; i < MAX_LINES; i++)
  {
    if (cur_state == lines[i].state)
    {
      cur_state = i;
      err_l = i;
      break;
    }
  }
  int error = 1;
  for (int i = 0; i < MAX_SYMVOLS; i++)
    if (alfabet[i] == cur_sym)
    {
      cur_move = lines[cur_state].symvols[i].command;
      cur_sym = lines[cur_state].symvols[i].symvol;
      cur_state = lines[cur_state].symvols[i].state;
      error = 0;
      err_c = i;
      break;
    }
  if (cur_state > MAX_LINES - 1)
  {
    sprintf(error_str, "Line %d, column %d:\n", err_l + 1, err_c + 1);
    Reset();
    strcat(error_str, "State out of range\n");
    MessageBox("Error", error_str, 0);
    return 0;
  }
  if (error)
  {
    sprintf(error_str, "Line %d, symvol '%c':\n", err_l + 1, cur_sym);
    Reset();
    strcat(error_str, "Symvol not found in\nalfabet\n");
    MessageBox("Error", error_str, 0);
    return 0;
  }
  if (cur_move == -1)
  {
    sprintf(error_str, "Line %d, column %d:\n", err_l + 1, err_c + 1);
    Reset();
    strcat(error_str, "Motion not specified\n");
    MessageBox("Error", error_str, 0);
    return 0;
  }
  if (cur_state == -1)
  {
    sprintf(error_str, "Line %d, column %d:\n", err_l + 1, err_c + 1);
    Reset();
    strcat(error_str, "New state not specified\n");
    MessageBox("Error", error_str, 0);
    return 0;
  }
  tape -> WriteData(cur_sym);
  if (cur_move == 0) tape -> MoveLeft();
  if (cur_move == 1) tape -> MoveRight();
  state -> Update(cur_step + 1, tape -> GetData(), cur_move, cur_state);
  if ((cur_move == 2) && (cur_sym == t) && (cur_state == s))
  {
    MessageBox("Information", "Program complete\n", 0);
    return 0;
  }
  if (lines[cur_state].is_break) return 0;
  cur_step++;
  return 1;
}

void TProgram::Reset()
{
  pos_x = 0;
  pos_y = 0;
  e1 = 0;
  e2 = 1;
  e3 = 0;
  cur_sym = ' ';
  cur_move = 0;
  cur_state = 0;
  state -> Update(-1, cur_sym, cur_move, cur_state);
  cur_step = 0;
  Update();
}

void TProgram::ClearBreakPoints()
{
  for (int i = 0; i < MAX_LINES; i++)
    lines[i].is_break = 0;
  Update();
}

void TProgram::GetState()
{
}

void TProgram::UpdateFilename()
{
  char buf[40];
  char *t = filename;
  strcpy(buf, " Program ");
  while (*t)
  {
    *t = toupper(*t);
    t++;
  }
  strcat(buf, filename);
  strcat(buf, " ");
  fill(11, 5, 23, 1, '', PROGRAM_COLOR);
  out_string(22 - strlen(buf) / 2, 5, buf, PROGRAM_TITLECOLOR);
}

void TProgram::Save()
{
  if (InputBox("Save program", "Enter file name:", 13, filename, filename, 0))
    return;
  FILE *out = fopen(filename, "wb");

  if (!out)
  {
    MessageBox("Error", "Cannot write file\n", 0);
    return;
  }

  fwrite(alfabet, MAX_SYMVOLS, 1, out);

  for (int i = 0; i < MAX_LINES; i++)
  {
    for (int j = 0; j < MAX_SYMVOLS; j++)
    {
      putc(lines[i].symvols[j].symvol, out);
      putc(lines[i].symvols[j].command & 255, out);
      putc(lines[i].symvols[j].state & 255, out);
    }
  }

  fclose(out);
  UpdateFilename();
}

void TProgram::Load()
{
  if (InputBox("Load program", "Enter file name:", 13, "*.TUB", filename, 0))
    return;
  FILE *in = fopen(filename, "rb");
  int t;

  if (!in)
  {
    MessageBox("Error", "Cannot open file\n", 0);
    return;
  }

  fread(alfabet, MAX_SYMVOLS, 1, in);

  for (int i = 0; i < MAX_LINES; i++)
  {
    for (int j = 0; j < MAX_SYMVOLS; j++)
    {
      t = getc(in);
      lines[i].symvols[j].symvol = t;
      t = getc(in);
      if (t == 255) t = -1;
      lines[i].symvols[j].command = t;
      t = getc(in);
      if (t == 255) t = -1;
      lines[i].symvols[j].state = t;
    }
  }

  fclose(in);
  Update();
  UpdateFilename();
}


#endif
