#ifndef _TYPE_H_
#define _TYPE_H_
#include <stdio.h>
#include <conio.h>
#include "defs.h"
#include "video.h"
#include "utils.h"
#include "clock.h"

class TTape {
    int pos;
    char data[1000];
    void Draw();
    void Update();
    char filename[40];
  public:
    TTape();
    ~TTape();
    int GetData();
    void WriteData(char);
    void Clear();
    void Edit(char, char);
    void MoveRight();
    void MoveLeft();
    void Save();
    void Load();
    int CheckEmpty();
};

TTape::TTape()
{
  strcpy(filename, "NONAME.TAP");
  pos = 500;
  Draw();
  Clear();
  gotoxy(39, 22);
}

TTape::~TTape()
{
}

int TTape::GetData()
{
  if (pos + 8 > 999) pos = 1000 - pos;
  return data[pos + 8];
}

void TTape::WriteData(char ch)
{
  if (pos + 8 > 999) pos = 1000 - pos;
  data[pos + 8] = ch;
  Update();
}

void TTape::Draw()
{
  fill(5, 19, 67, 1, '', TAPE_COLOR);
  fill(5, 21, 67, 1, '', TAPE_COLOR);
  fill(5, 20, 67, 1, ' ', TAPE_COLOR);
  out_char(4, 19, '', TAPE_COLOR);
  out_char(72, 19, '', TAPE_COLOR);
  out_char(4, 21, '', TAPE_COLOR);
  out_char(72, 21, '', TAPE_COLOR);
  out_char(4, 20, '', TAPE_COLOR);
  out_char(72, 20, '', TAPE_COLOR);
  set_attr(6, 22, 69, 1, SHADOW_COLOR);
  set_attr(73, 20, 2, 3, SHADOW_COLOR);
  for (int i = 0; i < 16; i++)
  {
    out_char(8 + i * 4, 19, '', TAPE_COLOR);
    out_char(8 + i * 4, 20, '', TAPE_COLOR);
    out_char(8 + i * 4, 21, '', TAPE_COLOR);
  }
  out_string(37, 21, "  ", TAPE_POINTCOLOR);
  Update();
}

void TTape::Update()
{
  for (int i = 0; i < 17; i++)
  {
    if (pos + i > 999)
      out_char(6 + i * 4, 20, data[pos + i - 1000], TAPE_CHARCOLOR);
    else
      out_char(6 + i * 4, 20, data[pos + i], TAPE_CHARCOLOR);
  }
}

void TTape::Clear()
{
  for (int i = 0; i < 1000; i++)
    data[i] = ' ';
  Update();
}

void TTape::Edit(char command, char key)
{
  gotoxy(39, 22);
  switch (command)
  {
    case 75: if (pos > 0) pos--; else pos = 999; break;
    case 77: if (pos < 1000) pos++; else pos = 0; break;
    case 83: key = 0x20; break;
  }
  if (key >= 0x20)
  {
    if (pos + 8 > 999)
      data[pos + 8 - 1000] = key;
    else
      data[pos + 8] = key;
  }
  if (key == 8)
  {
    if (pos + 7 > 999)
      data[pos + 7 - 1000] = ' ';
    else
      data[pos + 7] = ' ';
    pos--;
    if (pos < 0) pos = 999;
  }
  Update();
}

void TTape::MoveRight()
{
  pos++;
  if (pos > 999) pos = 0;
  Update();
}

void TTape::MoveLeft()
{
  pos--;
  if (pos < 0) pos = 999;
  Update();
}

int TTape::CheckEmpty()
{
  for (int i = 0; i < 1000; i++)
    if (data[i] != ' ') return 0;
  MessageBox("Warning", "Tape is empty\n", 0);
  return 1;
}

void TTape::Save()
{
  if (InputBox("Save tape", "Enter file name:", 13, filename, filename, 0))
    return;
  FILE *out = fopen(filename, "wb");

  if (!out)
  {
    MessageBox("Error", "Cannot write file\n", 0);
    return;
  }

  fwrite((char *)&pos, 2, 1, out);
  fwrite(data, 1000, 1, out);

  fclose(out);
}

void TTape::Load()
{
  if (InputBox("Load tape", "Enter file name:", 13, "*.TAP", filename, 0))
    return;
  FILE *in = fopen(filename, "rb");

  if (!in)
  {
    MessageBox("Error", "Cannot open file\n", 0);
    return;
  }

  fread((char *)&pos, 2, 1, in);
  fread(data, 1000, 1, in);

  fclose(in);
  Update();
}


#endif
